==> javie-0.3.1-src.tar.gz: javie-0.3.1-src/ch.kuramo.javie.core/src/main/java/ch/kuramo/javie/core/output/PullSourceOutput.java <==
/*
 * Copyright (c) 2010 Yoshikazu Kuramochi
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package ch.kuramo.javie.core.output;

import java.io.File;
import java.util.Arrays;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

import javax.media.opengl.GL;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.kuramo.javie.api.AudioMode;
import ch.kuramo.javie.api.RenderResolution;
import ch.kuramo.javie.api.Time;
import ch.kuramo.javie.core.AudioBuffer;
import ch.kuramo.javie.core.Composition;
import ch.kuramo.javie.core.VideoBuffer;
import ch.kuramo.javie.core.services.AudioRenderContext;
import ch.kuramo.javie.core.services.AudioRenderSupport;
import ch.kuramo.javie.core.services.VideoRenderContext;
import ch.kuramo.javie.core.services.VideoRenderSupport;

import com.google.inject.Inject;

public abstract class PullSourceOutput {

	private static final Logger _logger = LoggerFactory.getLogger(PullSourceOutput.class);


	protected static class VideoRequest {
		protected final Time time;
		protected VideoRequest(Time time) {
			this.time = time;
		}
	}

	protected static class AudioRequest {
		protected final Time time;
		protected final int frameCount;
		protected AudioRequest(Time time, int frameCount) {
			this.time = time;
			this.frameCount = frameCount;
		}
	}

	private static final VideoRequest END_VIDEO = new VideoRequest(null);

	private static final AudioRequest END_AUDIO = new AudioRequest(null, 0);


	private ProgressMonitor _progressMonitor;

	private final BlockingQueue<VideoRequest> _videoRequestQueue = new ArrayBlockingQueue<VideoRequest>(1);

	private final BlockingQueue<Object> _videoResponseQueue = new ArrayBlockingQueue<Object>(1);

	private final BlockingQueue<AudioRequest> _audioRequestQueue = new ArrayBlockingQueue<AudioRequest>(1);

	private final BlockingQueue<Object> _audioResponseQueue = new ArrayBlockingQueue<Object>(1);

	@Inject
	private VideoRenderContext _vrContext;

	@Inject
	private VideoRenderSupport _vrSupport;

	@Inject
	private AudioRenderContext _arContext;

	@Inject
	private AudioRenderSupport _arSupport;


	public void doOutput(final Composition comp, final AudioMode audioMode, File file, ProgressMonitor progMon) {
		Thread videoThread = new Thread(new Runnable() { public void run() { videoLoop(comp); } });
		videoThread.start();

		Thread audioThread = new Thread(new Runnable() { public void run() { audioLoop(comp, audioMode); } });
		audioThread.start();

		_progressMonitor = (progMon != null) ? progMon : ProgressMonitor.NULL;

		doOutput(comp, audioMode, file);

		try {
			_videoRequestQueue.put(END_VIDEO);
		} catch (InterruptedException e) {
			_logger.warn("unexpected interruption", e);
		}
		try {
			_audioRequestQueue.put(END_AUDIO);
		} catch (InterruptedException e) {
			_logger.warn("unexpected interruption", e);
		}
	}

	protected abstract void doOutput(Composition comp, AudioMode audioMode, File file);


	protected void videoLoop(Composition comp) {
		_vrContext.activate();
		try {
			GL gl = _vrContext.getGL();

			comp.prepareExpression(_vrContext.createInitialExpressionScope());

			while (true) {
				VideoBuffer vb = null;
				try {
					VideoRequest request = _videoRequestQueue.take();
					Time time = request.time;

					if (time == null) break;

					_vrContext.reset();
					_vrContext.setRenderResolution(RenderResolution.FULL);
					_vrContext.setTime(time);

					vb = comp.renderVideoFrame();

					// TODO アルファ無し／ストレートアルファ／乗算済みアルファを選択できるよにする
					_vrSupport.premultiply(vb);

					gl.glFinish();

					finishVideoRequest(request, vb);
					_videoResponseQueue.put(this);		// 入れる値は何でもよい。

				} catch (InterruptedException e) {
					_logger.warn("unexpected interruption", e);
				} finally {
					if (vb != null) vb.dispose();
				}
			}
		} finally {
			_vrContext.deactivate();
		}
	}

	protected void postVideoRequest(VideoRequest request) throws InterruptedException {
		_videoRequestQueue.put(request);
		_videoResponseQueue.take();
	}

	protected abstract void finishVideoRequest(VideoRequest request, VideoBuffer vb);


	protected void audioLoop(Composition comp, AudioMode audioMode) {
		long totalFrames = comp.getDuration().toFrameNumber(audioMode.sampleDuration);

		_arContext.activate();
		try {
			comp.prepareExpression(_arContext.createInitialExpressionScope());

			while (true) {
				AudioBuffer ab = null;
				try {
					AudioRequest request = _audioRequestQueue.take();
					Time time = request.time;

					if (time == null) break;

					_arContext.reset();
					_arContext.setAudioMode(audioMode);
					_arContext.setTime(time);
					_arContext.setFrameCount((int) Math.min(request.frameCount, totalFrames-time.timeValue));

					ab = comp.renderAudioChunk();
					if (ab == null) {
						ab = createAudioBufferClear();
					}

					finishAudioRequest(request, ab);
					_audioResponseQueue.put(this);		// 入れる値は何でもよい。

				} catch (InterruptedException e) {
					_logger.warn("unexpected interruption", e);
				} finally {
					if (ab != null) ab.dispose();
				}
			}
		} finally {
			_arContext.deactivate();
		}
	}

	protected void postAudioRequest(AudioRequest request) throws InterruptedException {
		_audioRequestQueue.put(request);
		_audioResponseQueue.take();
	}

	protected abstract void finishAudioRequest(AudioRequest request, AudioBuffer ab);


	private AudioBuffer createAudioBufferClear() {
		// このメソッドは AudioRenderSupport にあってもいいかもしれない。

		AudioBuffer ab = _arSupport.createAudioBuffer();

		switch (ab.getAudioMode().dataType) {
			case SHORT:
				Arrays.fill((short[])ab.getData(), 0, ab.getDataLength(), (short)0);
				break;
			case INT:
				Arrays.fill((int[])ab.getData(), 0, ab.getDataLength(), 0);
				break;
			case FLOAT:
				Arrays.fill((float[])ab.getData(), 0, ab.getDataLength(), 0);
				break;
			default:
				throw new UnsupportedOperationException(
						"unsupported AudioMode.DataType: " + ab.getAudioMode().dataType);
		}

		return ab;
	}


	protected void beginTask(int totalWork) {
		_progressMonitor.beginTask(totalWork);
	}

	protected void currentWork(int currentWork) {
		_progressMonitor.currentWork(currentWork);
	}

	protected void done() {
		_progressMonitor.done();
	}

	protected boolean isCanceled() {
		return _progressMonitor.isCanceled();
	}

}

==> javie-0.3.4-src.tar.gz: javie-0.3.4-src/ch.kuramo.javie.core/src/main/java/ch/kuramo/javie/core/output/PullSourceOutput.java <==
/*
 * Copyright (c) 2010 Yoshikazu Kuramochi
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package ch.kuramo.javie.core.output;

import java.io.File;
import java.util.Arrays;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

import javax.media.opengl.GL;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.kuramo.javie.api.AudioMode;
import ch.kuramo.javie.api.RenderResolution;
import ch.kuramo.javie.api.Time;
import ch.kuramo.javie.core.AudioBuffer;
import ch.kuramo.javie.core.Composition;
import ch.kuramo.javie.core.VideoBuffer;
import ch.kuramo.javie.core.services.AudioRenderContext;
import ch.kuramo.javie.core.services.AudioRenderSupport;
import ch.kuramo.javie.core.services.VideoRenderContext;
import ch.kuramo.javie.core.services.VideoRenderSupport;

import com.google.inject.Inject;

public abstract class PullSourceOutput {

	private static final Logger _logger = LoggerFactory.getLogger(PullSourceOutput.class);


	protected static class VideoRequest {
		protected final Time time;
		protected VideoRequest(Time time) {
			this.time = time;
		}
	}

	protected static class AudioRequest {
		protected final Time time;
		protected final int frameCount;
		protected AudioRequest(Time time, int frameCount) {
			this.time = time;
			this.frameCount = frameCount;
		}
	}

	private static final VideoRequest END_VIDEO = new VideoRequest(null);

	private static final AudioRequest END_AUDIO = new AudioRequest(null, 0);


	private ProgressMonitor _progressMonitor;

	private final BlockingQueue<VideoRequest> _videoRequestQueue = new ArrayBlockingQueue<VideoRequest>(1);

	private final BlockingQueue<Object> _videoResponseQueue = new ArrayBlockingQueue<Object>(1);

	private final BlockingQueue<AudioRequest> _audioRequestQueue = new ArrayBlockingQueue<AudioRequest>(1);

	private final BlockingQueue<Object> _audioResponseQueue = new ArrayBlockingQueue<Object>(1);

	@Inject
	private VideoRenderContext _vrContext;

	@Inject
	private VideoRenderSupport _vrSupport;

	@Inject
	private AudioRenderContext _arContext;

	@Inject
	private AudioRenderSupport _arSupport;


	public void doOutput(final Composition comp, final AudioMode audioMode, File file, ProgressMonitor progMon) {
		Thread videoThread = new Thread(new Runnable() { public void run() { videoLoop(comp); } });
		videoThread.start();

		Thread audioThread = new Thread(new Runnable() { public void run() { audioLoop(comp, audioMode); } });
		audioThread.start();

		_progressMonitor = (progMon != null) ? progMon : ProgressMonitor.NULL;

		doOutput(comp, audioMode, file);

		try {
			_videoRequestQueue.put(END_VIDEO);
		} catch (InterruptedException e) {
			_logger.warn("unexpected interruption", e);
		}
		try {
			_audioRequestQueue.put(END_AUDIO);
		} catch (InterruptedException e) {
			_logger.warn("unexpected interruption", e);
		}
	}

	protected abstract void doOutput(Composition comp, AudioMode audioMode, File file);


	protected void videoLoop(Composition comp) {
		_vrContext.activate();
		try {
			GL gl = _vrContext.getGL();

			comp.prepareExpression(_vrContext.createInitialExpressionScope(comp));

			while (true) {
				VideoBuffer vb = null;
				try {
					VideoRequest request = _videoRequestQueue.take();
					Time time = request.time;

					if (time == null) break;

					_vrContext.reset();
					_vrContext.setRenderResolution(RenderResolution.FULL);
					_vrContext.setTime(time);

					vb = comp.renderVideoFrame();

					// TODO アルファ無し／ストレートアルファ／乗算済みアルファを選択できるよにする
					_vrSupport.premultiply(vb);

					gl.glFinish();

					finishVideoRequest(request, vb);
					_videoResponseQueue.put(this);		// 入れる値は何でもよい。

				} catch (InterruptedException e) {
					_logger.warn("unexpected interruption", e);
				} finally {
					if (vb != null) vb.dispose();
				}
			}
		} finally {
			_vrContext.deactivate();
		}
	}

	protected void postVideoRequest(VideoRequest request) throws InterruptedException {
		_videoRequestQueue.put(request);
		_videoResponseQueue.take();
	}

	protected abstract void finishVideoRequest(VideoRequest request, VideoBuffer vb);


	protected void audioLoop(Composition comp, AudioMode audioMode) {
		long totalFrames = comp.getDuration().toFrameNumber(audioMode.sampleDuration);

		_arContext.activate();
		try {
			comp.prepareExpression(_arContext.createInitialExpressionScope(comp));

			while (true) {
				AudioBuffer ab = null;
				try {
					AudioRequest request = _audioRequestQueue.take();
					Time time = request.time;

					if (time == null) break;

					_arContext.reset();
					_arContext.setAudioMode(audioMode);
					_arContext.setTime(time);
					_arContext.setFrameCount((int) Math.min(request.frameCount, totalFrames-time.timeValue));
					_arContext.setEvaluationResolution(audioMode.sampleRate/100);

					ab = comp.renderAudioChunk();
					if (ab == null) {
						ab = createAudioBufferClear();
					}

					finishAudioRequest(request, ab);
					_audioResponseQueue.put(this);		// 入れる値は何でもよい。

				} catch (InterruptedException e) {
					_logger.warn("unexpected interruption", e);
				} finally {
					if (ab != null) ab.dispose();
				}
			}
		} finally {
			_arContext.deactivate();
		}
	}

	protected void postAudioRequest(AudioRequest request) throws InterruptedException {
		_audioRequestQueue.put(request);
		_audioResponseQueue.take();
	}

	protected abstract void finishAudioRequest(AudioRequest request, AudioBuffer ab);


	private AudioBuffer createAudioBufferClear() {
		// このメソッドは AudioRenderSupport にあってもいいかもしれない。

		AudioBuffer ab = _arSupport.createAudioBuffer();

		switch (ab.getAudioMode().dataType) {
			case SHORT:
				Arrays.fill((short[])ab.getData(), 0, ab.getDataLength(), (short)0);
				break;
			case INT:
				Arrays.fill((int[])ab.getData(), 0, ab.getDataLength(), 0);
				break;
			case FLOAT:
				Arrays.fill((float[])ab.getData(), 0, ab.getDataLength(), 0);
				break;
			default:
				throw new UnsupportedOperationException(
						"unsupported AudioMode.DataType: " + ab.getAudioMode().dataType);
		}

		return ab;
	}


	protected void beginTask(int totalWork) {
		_progressMonitor.beginTask(totalWork);
	}

	protected void currentWork(int currentWork) {
		_progressMonitor.currentWork(currentWork);
	}

	protected void done() {
		_progressMonitor.done();
	}

	protected boolean isCanceled() {
		return _progressMonitor.isCanceled();
	}

}
